require 'json'
require 'base64'
require 'uu/os/lang/json_patch'
require 'uu/os/ipc/ipc_error'
require 'uu/os/application/application_error'
require 'uu/os/cmd/command_error'
require 'uu/os/cmd/command_future_actor'
require 'uu/os/lang/future'

module UU
  module OS
    module CMD

      # Default command handler. As custom handler might be used any class (or object instance)
      # which responds to at least one of methods defined in this class.
      class CommandHandler

        # Value to be returned by handler in case it does not know
        # how to process given values and therefore processing should
        # be delegated to next handler.
        NEXT = :__next__

        # Allows to transform given payload as needed along with related modification of HTTP request headers and query.
        # @param [Object] payload Original request payload.  Note that in case payload is IO (or structure containing IO)
        #    it might not be available to next handlers in chain once consumed (depends on support of :rewind operation).
        # @param [UU::OS::HTTP::HTTPHeaders] http_headers Modifiable original request headers.
        # @param [UU::OS::HTTP::HTTPQuery] http_parameters Modifiable original request query parameters.
        # @param [Hash] options Additional options given for request (see {UU::OS::CMD::CommandClient#invoke
        #    CommandClient#invoke} for details.)
        # @return [Object, Symbol, Exception] Transformed payload (including nil) or value of {NEXT NEXT} to pass
        #    processing to next handler. In case method returns instance of exception, it will be immediately raised.
        def handle_request(payload, http_headers, http_parameters, options)
          invocation_method = options[:invocation_method]
          if invocation_method == :get
            if payload.nil? || payload.kind_of?(Hash)
              # Let transformation of query parameters on HTTPClient
              return payload
            else
              raise TypeError, "Cannot transform #{payload.class} to query parameters."
            end
          elsif invocation_method == :post
            return payload if http_headers.contains(:content_type)
            binary_upload = options[:binary_upload] || false
            multipart_upload = options[:multipart_upload] || false
            if binary_upload
              http_headers[:content_type] = 'multipart/form-data'
              if payload.kind_of?(Hash)
                bmp = payload.dup
                # Ensure data are set as last parameter
                bmp[:data] = bmp.delete(:data)
                # Let transformation of multipart on HTTPClient
                return bmp
              elsif payload.kind_of?(String) || !payload.kind_of?(IO) || !payload.kind_of?(StringIO) || !payload.kind_of?(Tempfile)
                bmp = {}
                bmp[:name] = (payload.name if payload.respond_to?(:name)) || (File.basename(payload) if payload.respond_to?(:path))
                bmp[:contentType] = payload.content_type if payload.respond_to?(:content_type)
                bmp[:size] = (payload.size if payload.respond_to?(:size)) || (payload.length if payload.respond_to?(:length))
                bmp[:data] = payload
                # Let transformation of multipart on HTTPClient
                return bmp
              else
                raise TypeError, "Cannot transform #{payload.class} to binary message."
              end
            elsif multipart_upload
              http_headers[:content_type] = 'multipart/form-data'
              if payload.nil? || payload.kind_of?(Hash)
                # Let transformation of multipart on HTTPClient
                return payload
              else
                raise TypeError, "Cannot transform #{payload.class} to multipart message."
              end
            else
              http_headers[:content_type] = 'application/json'
              # Let transformation of JSON on HTTPClient
              return payload
            end
          else
            raise NotImplementedError, "#{options[:invocation_method].upcase} commands are not supported."
          end
        end

        # Allows processing of raw response in order to return expected result.
        # @param [UU::OS::HTTP::HTTPResponse] http_response Raw command response (read only). Note that in case
        #    response body is IO, once it is consumed it will not be available to next handlers in chain.
        # @param [Hash] options Additional options given for request (see {UU::OS::CMD::CommandClient#invoke
        #    CommandClient#invoke} for details.)
        # @return [Object, Symbol, Exception] Processed command response (including nil) or value of {NEXT NEXT} to pass
        #    processing to next handler. In case method returns instance of exception, it will be immediately raised.
        def handle_response(http_response, options)
          if http_response.status == 204
            return nil
          elsif options[:raw_response]
            # Do not touch body if raw response was requested
            return http_response.body
          elsif http_response.headers.contains(:content_disposition, 'attachment')
            # Legacy commands may send additional metadata header
            metadata = http_response.headers[:uu_os_binary_data_info]
            metadata &&= JSON.parse(Base64.decode64(metadata[0]), symbolize_names: true)
            metadata ||= {}
            # Transform file/binary download into binary data structure
            binary = {name: nil, content_type: nil, size: nil, data: nil}
            if metadata.has_key?(:name) || metadata.has_key?(:fileName)
              binary[:name] = metadata[:name] || metadata[:fileName]
            elsif http_response.headers[:content_disposition][0] =~ /filename=([^;]+)/
              name = $1
              name = name[1..-2] if name =~ /^".*"$/
              binary[:name] = name
            end
            if metadata.has_key?(:contentType)
              binary[:content_type] = metadata[:contentType]
            elsif !http_response.headers[:content_type].nil?
              binary[:content_type] = http_response.headers[:content_type][0]
            end
            if metadata.has_key?(:size)
              binary[:size] = metadata[:size]
            elsif !http_response.headers[:content_length].nil?
              binary[:size] = http_response.headers[:content_length][0].to_i
            end
            data = http_response.body
            # Binary data structure should always return data as IO
            if data.kind_of?String
              binary[:data] = StringIO.new(data)
            else
              binary[:data] = data
            end
            return binary
          elsif options[:stream_response]
            # If stream was requested, it must be returned
            body = http_response.body
            if body.kind_of?String
              return StringIO.new(body)
            else
              return body
            end
          elsif http_response.headers.contains(:content_type, 'application/json')
            # In case content type is JSON, we should parse it
            json = http_response.body
            json = json.read if json.kind_of?IO
            json.force_encoding('UTF-8')
            if json.empty? || json =~ /^{"nil":"true"}$/i
              # Second condition is null serialized by JAXB
              result = nil
            elsif json =~ /^\".*\"$/
              # String value 
              return json[1..-2]
            else
              # Parse JSON (leave original value if parse is not possible)
              result = (JSON.parse(json, symbolize_names: true, quirks_mode: true) rescue json)
            end
            if UU::OS::CMD::CommandFutureActor.command_future?(result)
              actor = UU::OS::CMD::CommandFutureActor.new(result, options)
              return UU::OS::Lang::Future.new(actor)
            else
              return result
            end
          else
            # Don't know how to handle, custom handler implementation is needed.
            return http_response.body
          end
        end

        # Allows processing of command invocation error.
        # @param [UU::OS::HTTP::HTTPResponse] http_response Raw command response (read only). Note that in case
        #    response body is IO, once it is consumed it will not be available to next handlers in chain.
        # @param [Hash] options Additional options given for request (see {UU::OS::CMD::CommandClient#invoke
        #    CommandClient#invoke} for details.)
        # @return [Object, Symbol, Exception] Processed command error (including nil) or value of {NEXT NEXT} to pass
        #    processing to next handler. In case method returns instance of exception, it will be immediately raised.
        def handle_error(http_response, options)
          body = http_response.body
          body = body.read if body.respond_to?:read
          begin
            body = JSON.parse(body, symbolize_names: true)
            if body[:code] =~ /^UU.OS.IPC\//
              raise UU::OS::IPC::IPCError.new(body)
            elsif body[:code] =~ /^UU.OS.APPLICATION\//
              raise UU::OS::Application::ApplicationError.new(body)
            else
              raise UU::OS::CMD::CommandError.new(body)
            end
          rescue JSON::ParserError
            # In case body is not JSON, it is not exception raised by
            # our framework, but exception raised by infrastructure
            # (therefore it can be considered as IPC error).
            raise UU::OS::IPC::IPCError.new(body)
          end
        end

        # Allows processing of HTTP redirection.
        # @param [UU::OS::HTTP::HTTPResponse] http_response Raw command response (read only). Note that in case
        #    response body is IO, once it is consumed it will not be available to next handlers in chain.
        # @param [Hash] options Additional options given for request (see {UU::OS::CMD::CommandClient#invoke
        #    CommandClient#invoke} for details.)
        # @return [Object, Symbol, Exception] Result of processed redirection (including nil) or value of {NEXT NEXT}
        #    to pass processing to next handler. In case method returns instance of exception, it will be immediately
        #    raised.
        def handle_redirect(http_response, options)
          raise NotImplementedError, "Service #{http_response.request.url} responded with #{http_response.status}, but command redirection is not supported."
        end

      end

    end
  end
end
